from utilities.solutionTS import Solution
from utilities.transformation import *
import copy
from scipy.special import gamma
import numpy as np

class HHO:
    def __init__(self, N, function):
        self.N = N
        self.population = []
        self.function = function
        self.rabbit_location = [None] * self.function.D # best solution
        self.FFE= self.N

    
    def initial_population(self):
        for i in range(self.N):
            local_solution = Solution(self.function)
            self.population.append(local_solution)
            
        self.population.sort(key=lambda x: x.objective_function)
        self.rabbit_location = copy.deepcopy(self.population[0].x)

    def update_position(self, t, max_iter):
        
        Xbest = np.array(copy.deepcopy(self.rabbit_location))
        
        pop = []
        for i in range(self.N):
            pop.append(self.population[i].x)
            
        Xmean = np.mean(np.array(pop))
            
        lb = self.function.lb
        ub = self.function.ub
        
        for i in range(self.N):
      
            # Update the initial energy E0 and jump strength J E0=2rand()-1, J=2(1-rand()) Update the E using Eq. (3)
            E0 = 2 * np.random.uniform(0, 1) - 1 
            J = 2 * (1 - np.random.uniform(0, 1))
            E = 2 * E0 * (1 - (t/max_iter)) # Escaping energy
            Xcurr = np.array(copy.deepcopy(self.population[i].x))
            if(np.abs(E) >= 1): # Exploration phase
                # Update the location vector using Eq. (1)
                q = np.random.uniform(0,1)
                if (q >= 0.5):
                    Xrand = np.array(copy.deepcopy(self.population[np.random.randint(0,self.N)].x))
                    r1 = np.random.uniform(0, 1) 
                    r2 = np.random.uniform(0, 1) 
                    Xnew = Xrand - r1* np.abs(Xrand - 2 * r2 * Xcurr)
                else:
                    # Xm is the average position of the current population of hawks
                    r3 = np.random.uniform(0, 1) 
                    r4 = np.random.uniform(0, 1)
                    Xnew = Xbest - Xmean - r3 * (lb[0] + r4 * (ub[0] - lb[0]))
                    
            else: # Exploitation phase
                if np.random.uniform(0,1) >= 0.5: 
                    Xdelta = Xbest - Xcurr
                    # Update the location vector using Eq. (4)
                    if np.abs(E) >= 0.5: # Soft besiege
                        Xnew = Xdelta - E * np.abs(J * Xbest - Xcurr)
                    # Update the location vector using Eq. (6)
                    else: # Hard besiege
                        Xnew = Xbest - E * np.abs(Xdelta)

                    Xnew = np.array(Xnew)

                    # konvertujemo sve elmente koji treba da budu integer u integer
                    Xnew = convertToInt(Xnew, self.function.intParams)

                    # Xnew = convertToInt(Xnew, self.function.intParams)

                    Xnew = self.checkBoundaries(Xnew)

                    Xnew = convertToInt(Xnew, self.function.intParams)

                    self.FFE = self.FFE + 1
                    solution = Solution(self.function, Xnew)

                    if solution.objective_function<self.population[i].objective_function:
                        self.population[i] = solution
                        
                else:
                    sigma = np.power((gamma(1 + 1.5) * np.sin(np.pi * 1.5 / 2.0)) / (gamma((1 + 1.5) * 1.5 * np.power(2, (1.5 - 1) / 2)) / 2.0), 1.0 / 1.5)
                    # levy flight function
                    LF = 0.01 * np.random.uniform(0,1) * sigma / np.power(np.abs(np.random.uniform(0,1)), 1.0 / 1.5)
                    # S is a random vector by size 1 × D
                    S = np.random.uniform(lb[0], ub[0], self.function.D)
                    
                    if (np.abs(E) >= 0.5): # Soft besiege with progressive rapid dives
                        # Update the location vector using Eq. (10)
                        Y = Xbest - E * np.abs( J * Xbest - Xcurr)
                        Y = np.array(Y)

                        # konvertujemo sve elmente koji treba da budu integer u integer
                        Y = convertToInt(Y, self.function.intParams)
                        Y = self.checkBoundaries(Y)
                        Y_solution = Solution(self.function,  Y.tolist())
                        # Generate a new solution
                        Y_fitness = Y_solution.objective_function
                    else: # Hard besiege with progressive rapid dives
                        # Update the location vector using Eq. (11)
                        Y = Xbest - E * np.abs(J * Xbest - Xmean)
                        Y = np.array(Y)

                        # konvertujemo sve elmente koji treba da budu integer u integer
                        Y = convertToInt(Y, self.function.intParams)
                        Y = self.checkBoundaries(Y)
                        Y_solution = Solution(self.function, Y.tolist())
                        # Generate a new solution
                        Y_fitness = Y_solution.objective_function
                    
 
                    Y = Xbest - E * np.abs( J * Xbest - Xcurr)
                    Z = Y + S * LF
                    Z = np.array(Z)

                    # konvertujemo sve elmente koji treba da budu integer u integer
                    Z = convertToInt(Z, self.function.intParams)
                    Z = self.checkBoundaries(Z)
                    Z_solution = Solution(self.function, Z.tolist())

                    Z_fitness = Z_solution.objective_function
                    
                    if Y_fitness < self.population[i].objective_function:
                        self.population[i] = Y_solution
                        
                    if Z_fitness < self.population[i].objective_function:
                        self.population[i] = Z_solution
        self.sort_population()
            
    def sort_population(self):

        self.population.sort(key=lambda x: x.objective_function)
        self.rabbit_location = self.population[0].x

    def get_global_best(self):
        return (self.population[0].objective_function, self.population[0].overallResults, self.population[0].allResults,
                self.population[0].overallResults1, self.population[0].allResults1, self.population[0].results,
                self.population[0].model)

    def get_global_worst(self):
        return (
        self.population[-1].objective_function, self.population[-1].overallResults, self.population[-1].allResults,
        self.population[-1].overallResults1, self.population[-1].allResults1, self.population[-1].results,
        self.population[-1].model)
    
    def optimum(self):
        print('f(x*) = ', self.function.minimum, 'at x* = ', self.function.solution)
        
    def algorithm(self):
        return 'HHO'
    
    def objective(self):
        
        result = []
        
        for i in range(self.N):
            result.append(self.population[i].objective_function)
            
        return result
    
    def average_result(self):
        return np.mean(np.array(self.objective()))
    
    def std_result(self):        
        return np.std(np.array(self.objective()))
    
    def median_result(self):
        return np.median(np.array(self.objective()))
        
       
    def print_global_parameters(self):
            for i in range(0, len(self.rabbit_location)):
                 print('X: {}'.format(self.rabbit_location[i]))
                 
    def get_best_solutions(self):
        return np.array(self.rabbit_location)

    def get_solutions(self):
        
        sol = np.zeros((self.N, self.function.D))
        for i in range(len(self.population)):
            sol[i] = np.array(self.population[i].x)
        return sol
    
    def print_all_solutions(self):
        print("******all solutions objectives**********")
        for i in range(0,len(self.population)):
              print('solution {}'.format(i))
              print('objective:{}'.format(self.population[i].objective_function))
              print('solution {}: '.format(self.population[i].x))
              print('--------------------------------------')

    def get_global_best_params(self):
        return self.population[0].x


    def checkBoundaries(self,Xnew):
        for j in range(self.function.D):
            if Xnew[j] < self.function.lb[j]:
                Xnew[j] = self.function.lb[j]

            elif Xnew[j] > self.function.ub[j]:
                Xnew[j] = self.function.ub[j]
        return Xnew

    def getFFE(self):
        #metoda za prikaz broja FFE
        return self.FFE

        # funkcija koja vraca najbolji global_best_solution

    # funkcija koja vraca najbolji global_best_solution
    # funkcija koja vraca najbolji global_best_solution, ali se cuva i diversifikacija u poislednjoj populaciji
    def get_global_best_solution(self):
        # ovde pravimo liste sa objective i indicator za celu populaciji

        indicator_list = []  # ovo je indikator, sta god da je u pitanju
        objective_list = []  # ovo je objective, sta god da je u pitanju
        objective_indicator_list = []
        for i in range(len(self.population)):
            indicator_list.append(self.population[
                                      i].overallResults[
                                      0])  # ovo je za indikator, u ovom slucaju R2, ali moze da se menja
            objective_list.append(self.population[i].objective_function)  # ovo je objective, sta god da je
        objective_indicator_list.append(objective_list)
        objective_indicator_list.append(indicator_list)
        self.population[0].diversity = objective_indicator_list

        return self.population[0]